/*
 *  IBM disk device driver
 *
 *  (c) 1998, Yoav Shadmi
 *
 */

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"

#include "IBM-DISK.h"

#include <go32.h>
#include <dpmi.h>
#include <bios.h>

// Floppy disk device driver [BIOS implementation]

int Disk_Check (void)
{
   struct diskinfo_t DiskHandle;
   char   dcheck_buf[512];
   int    _result;

   // We'll be doing this test on the last sector on the disk,
   // which is unused anyway..

   DiskHandle.drive    = 0;           // Drive A:
   DiskHandle.head     = 0;           //          
   DiskHandle.track    = 0;           // location 
   DiskHandle.sector   = 1;           //          
   DiskHandle.nsectors = 1;           // Number of sectors to read
   DiskHandle.buffer   = dcheck_buf;

   _result = (_bios_disk(_DISK_READ, &DiskHandle) & 0xff00) >> 8;
   _result = (_bios_disk(_DISK_READ, &DiskHandle) & 0xff00) >> 8;
   _result = (_bios_disk(_DISK_READ, &DiskHandle) & 0xff00) >> 8;
   _result = (_bios_disk(_DISK_WRITE, &DiskHandle) & 0xff00) >> 8;
   _result = (_bios_disk(_DISK_WRITE, &DiskHandle) & 0xff00) >> 8;
   _result = (_bios_disk(_DISK_WRITE, &DiskHandle) & 0xff00) >> 8;

   return _result;
}

int Disk_Read (char *buffer, int size, int drive, int LBN)
{
   struct diskinfo_t DiskHandle;
   int    _result, x=0, i;
   int    tLBN=LBN, tSize=size;

   if (((tSize/512)+(((tLBN/512)% 18)+1))<=19) {
      DiskHandle.drive    = 0;
      DiskHandle.nsectors = tSize/512;
      DiskHandle.sector   = ((tLBN/512)%18)+1;
      DiskHandle.head     = ((tLBN/512)/18)%2;
      DiskHandle.track    = ((tLBN/512)/36);
      DiskHandle.buffer   = (char *) buffer+x;

      _result = (_bios_disk(_DISK_READ, &DiskHandle) & 0xff00) >> 8;

      if (_result)
         return 255;

   } else {
      DiskHandle.drive    = 0;
      DiskHandle.nsectors = 1;

      for (i=0; i<(tSize/512); i++) {
         DiskHandle.sector   = ((tLBN/512)%18)+1;
         DiskHandle.head     = ((tLBN/512)/18)%2;
         DiskHandle.track    = ((tLBN/512)/36);
         DiskHandle.buffer   = (char *) buffer+x;
         
         _result = (_bios_disk(_DISK_READ, &DiskHandle) & 0xff00) >> 8;

         if (_result) {
            return 255;
         } else {
            x+=512;
            tLBN+=512;
         }
      }
   }

   return 0;
}

int Disk_Write (char *buffer, int size, int drive, int LBN)
{
   struct diskinfo_t DiskHandle;
   int    _result, x=0, i;
   int    tLBN=LBN, tSize=size;

   if (((tSize/512)+(((tLBN/512)% 18)+1))<=19) {
      DiskHandle.drive    = 0;
      DiskHandle.nsectors = tSize/512;
      DiskHandle.sector   = ((tLBN/512)%18)+1;
      DiskHandle.head     = ((tLBN/512)/18)%2;
      DiskHandle.track    = ((tLBN/512)/36);
      DiskHandle.buffer   = (char *) buffer+x;

      _result = (_bios_disk(_DISK_WRITE, &DiskHandle) & 0xff00) >> 8;

      if (_result)
         return 255;

   } else {
      DiskHandle.drive    = 0;
      DiskHandle.nsectors = 1;

      for (i=0; i<(tSize/512); i++) {
         DiskHandle.sector   = ((tLBN/512)%18)+1;
         DiskHandle.head     = ((tLBN/512)/18)%2;
         DiskHandle.track    = ((tLBN/512)/36);
         DiskHandle.buffer   = (char *) buffer+x;
         
         _result = (_bios_disk(_DISK_WRITE, &DiskHandle) & 0xff00) >> 8;

         if (_result) {
            return 255;
         } else {
            x+=512;
            tLBN+=512;
         }
      }
   }

   return 0;
}

void Disk_StopMotor (void)
{
   // stop the floppy disk motor, so that vMac resumes control at a known
   // machine state.

   outportb(0x03F2, 0x00);

   return; 
}

void Disk_LockDriver (void)
{
   _go32_dpmi_lock_code (Disk_Check,     (unsigned int)Disk_Read       - (unsigned int)Disk_Check);
   _go32_dpmi_lock_code (Disk_Read,      (unsigned int)Disk_Write      - (unsigned int)Disk_Read);
   _go32_dpmi_lock_code (Disk_Write,     (unsigned int)Disk_StopMotor  - (unsigned int)Disk_Write);
   _go32_dpmi_lock_code (Disk_StopMotor, (unsigned int)Disk_LockDriver - (unsigned int)Disk_StopMotor);

   return;
}
