/*
 *  IBM Printer/Modem device drivers
 *
 *  (c) 1998, Yoav Shadmi
 *
 *  The modem device driver is a stub, don't use it.
 */

#include "sysconfi.h"
#include "sysdeps.h"
#include "config.h"

#include "IBM-SCC.h"

#include <dos.h>
#include <go32.h>
#include <dpmi.h>

// Printer port driver [32bit Implementation]

#define rPAR_Data       0
#define rPAR_Status     1
#define rPAR_Control    2

#define Wait_Count      5715
#define TimeOut_Count   5

#define Wait(Count)                                              \
{                                                                \
   UWORD Wait_Counter;                                           \
                                                                 \
   for (Wait_Counter=0; Wait_Counter<Count; Wait_Counter++);     \
}

PAR_HANDLE Printer_InitPort (UWORD Port_Number)
{
   PAR_HANDLE PortHandle;
   UWORD Port=0xff;

   switch (Port_Number) {
      case 1: _dosmemgetw(0x408, 1, &Port); break;
      case 2: _dosmemgetw(0x40A, 1, &Port); break;
      case 3: _dosmemgetw(0x40C, 1, &Port); break;
   }

   // Check for printer port availability
   outportb(Port+rPAR_Data, 0xAA);
   Wait(Wait_Count)                      // Port wait macro       
   if (inportb(Port+rPAR_Data)!=0xAA) {
      PortHandle.PortNumber = 0xff;      // Port unavailable
   } else {                              
      outportb(Port+rPAR_Data, 0x00);    // Initialize port
      outportb(Port+rPAR_Control, 0x09); 
      Wait(Wait_Count)                   // Port wait macro       
      outportb(Port+rPAR_Control, 0x0C);
      if (inportb(Port+rPAR_Status)==0) {
         PortHandle.PortNumber=0xff;     // Initialization error
      } else {
         PortHandle.PortNumber=Port;     // Initialization successful
         PortHandle.PortStatus.PrinterStatus=inportb(Port+rPAR_Status);
      }
   }

   return PortHandle;
}

int Printer_WritePort (PAR_HANDLE PortHandle, UBYTE character)
{
   UWORD TimeOut=0;

   // Assume that the port is valid at this point
   outportb(PortHandle.PortNumber+rPAR_Data, character);
   outportb(PortHandle.PortNumber+rPAR_Control, 0x0D);
   Wait(Wait_Count)
   outportb(PortHandle.PortNumber+rPAR_Control, 0x0C);

   while (((inportb(PortHandle.PortNumber+rPAR_Status) & 0x80)==0) && (TimeOut<TimeOut_Count)) {
      TimeOut++;
   }

   PortHandle.PortStatus.PrinterStatus=inportb(PortHandle.PortNumber+rPAR_Status);

   if (TimeOut==5)      // Failed to write to the printer port
      return 1;

   return 0;
}

void Printer_GetPort (PAR_HANDLE PortHandle)
{
   PortHandle.PortStatus.PrinterStatus=inportb(PortHandle.PortNumber+rPAR_Status);

   return;
}

void Printer_LockDriver (void)
{
   return;
}

// Modem port driver
SER_HANDLE Modem_InitPort (UWORD Port_Number, UWORD BaudRate, UBYTE Parity, UBYTE StopBits, UBYTE DataBits)
{
   SER_HANDLE PortHandle;

   PortHandle.PortNumber = 0xff;

   return PortHandle;
}

int Modem_WritePort (SER_HANDLE PortHandle, UBYTE character)
{
   return 1;
}

UBYTE Modem_ReadPort (SER_HANDLE PortHandle, int result)
{
   result = 1;
   return 0x80;
}

void Modem_GetPort (SER_HANDLE PortHandle)
{
   PortHandle.PortStatus.LineStatus  = 0x80;
   PortHandle.PortStatus.ModemStatus = 0x00;

   return;
}

void Modem_ReinitPort (SER_HANDLE PortHandle)
{
   PortHandle = Modem_InitPort(PortHandle.PortNumber, PortHandle.PortParams.BaudRate, PortHandle.PortParams.Parity, PortHandle.PortParams.StopBits, PortHandle.PortParams.DataBits);
}

void Modem_LockDriver (void)
{
   return;
}
