/*
   vMac - Screen.C 

   (c) Gustavo Goedert, 1997
*/

#include "sysconfi.h"
#include "sysdeps.h"
#include "config.h"

#include "Screen.h"

#include "IBM-UI.h"
#include "IBM-VGA.h"
#include "IBM-TIME.h"

#include <sys/fsext.h>
#include <go32.h>
#include <dpmi.h>


#define MAXMESSAGES 100
char *Messages[MAXMESSAGES];
int CurrentMessage = 0;
int no_msg=0;

// Prototypes.
extern void m68k_reset  (void);
extern void Quit_vMac   (void);
extern void MacReset    (void);
static int  ScreenWrite (__FSEXT_Fnumber func, int *retval, va_list rest_args);

int Screen_Init(int argc, char **argv)
{
   int i;
   __dpmi_regs Regs;

   /* whenever we try to exit the application, Quit_Emulator will be called */
   atexit(Quit_Emulator);

   /* redirect output */
   __FSEXT_set_function(fileno (stdout), ScreenWrite);
   __FSEXT_set_function(fileno (stderr), ScreenWrite);
   for (i=0; i<MAXMESSAGES; i++)
      Messages[i] = NULL;

   /* Save Old VGA Mode */
   Regs.x.ax=0x1C;
   Regs.h.ah=0x0F;
   __dpmi_int(0x10, &Regs);
   OldVGAMode = Regs.h.al;

   /* start MS-DOS event handling. */
   if (!Mouse_Install())
      return 1;

   if (!VGA_InstallGraphics())
      return 1;

   return 0;
}

void Screen_Reset(void)
{
}

void DumpMessages(void)
{
   int i, StartMessage, MessagePosition;

   UseScreen = 1;
   if (CurrentMessage) {
      StartMessage = CurrentMessage - MAXMESSAGES;
      if (StartMessage < 0)
         StartMessage = 0;

      for (i=StartMessage; i<CurrentMessage; i++) {
         MessagePosition = i % MAXMESSAGES;
         if (Messages[MessagePosition] != NULL)
            printf(Messages[MessagePosition]);
      }
      CurrentMessage = 0;
   }
}

static int ScreenWrite(__FSEXT_Fnumber func, int *retval, va_list rest_args)
{
   char *buf, *mybuf;
   size_t buflen;
   int fd = va_arg (rest_args, int);
   int MessagePosition;

   if (func != __FSEXT_write || UseScreen)
      return 0;

   buf = va_arg(rest_args, char *);
   buflen = va_arg(rest_args, size_t);
   mybuf = malloc(buflen + 1);

   if (mybuf != NULL) {
      memcpy (mybuf, buf, buflen);
      mybuf[buflen] = '\0';
   }

   MessagePosition = CurrentMessage % MAXMESSAGES;
   if (Messages[ MessagePosition ] != NULL)
      free(Messages[ MessagePosition ]);

   Messages[ MessagePosition ] = mybuf;

   CurrentMessage++;

   *retval = buflen;
   return 1;
}

void Quit_Emulator (void)
{
   extern allegro_exit (void);

   if (no_msg)
      exit(1);

   Mouse_Uninstall();
   Timer_Uninstall();
   Keyboard_Uninstall();
   VGA_UninstallGraphics();

   allegro_exit();

   Quit_vMac();
}

void Reset_Emulator (void)
{
   MacReset();
}

// VIA Interface Functions

UBYTE VIA_GORA6 (void) // Main/Alternate Screen Buffer
{
#ifdef _VIA_Interface_Debug
   printf("VIA ORA6 attempts to be an input\n");
#endif
   return 0;
}

void  VIA_PORA6 (UBYTE Data)
{
   vPage2 = Data;
}

UBYTE VIA_GORB6 (void) // Video Beam in Display
{
   return 0; // Assume it is
}

void  VIA_PORB6 (UBYTE Data)
{
#ifdef _VIA_Interface_Debug
  printf("VIA ORB6 attempts to be an output\n");
#endif
}
