//
// vMac 0.1 Source
// By Philip Cummins (PC)
// Copyright PC 1997
//  

// NCR5380 chip emulation by Yoav Shadmi, 1998

#include "sysconfi.h"                   // includes/sysconfig.h
#include "sysdeps.h"
#include "config.h"

#include "cpu/machdep/m68k.h"
#include "memory2.h"                    // includes/memory.h
#include "Memory.h"
#include "custom2.h"                    // includes/vMac_custom.h
#include "patchmgr.h"

#include "readcpu.h"
#include "uaepatch.h"                   // includes/uae_vmac_patch.h
#include "newcpu.h"

#include "SCSI.h"

#ifdef _SCSI_Debug
#include <stdio.h>
#endif

UWORD private_SCSI [kSCSI_Size/2];
UWORD *SCSI = private_SCSI;

UBYTE CheckSCSI=0;
static calls=0;

SCSI_Ty SCSI_Chip;

void SCSI_ChipReset(void);
void SCSI_BusReset(void);
void SCSI_CheckPins (void);

UBYTE SCSI_GetB2 (CPTR addr);
UWORD SCSI_GetW2 (CPTR addr);
ULONG SCSI_GetL2 (CPTR addr);
void  SCSI_PutB2 (CPTR addr, UBYTE in);
void  SCSI_PutW2 (CPTR addr, UWORD in);
void  SCSI_PutL2 (CPTR addr, ULONG in);

int SCSI_Init (void)
{
   SCSI_Reset();
   return 0;
}

// SCSI Bus reset 
void SCSI_Reset (void)
{
   return;
}

void SCSI_ChipReset (void)
{
   SCSI_PutB2(scsiRd+sCDR, 0);
   SCSI_PutB2(scsiWr+sODR, 0);
   SCSI_PutB2(scsiRd+sICR, 0);
   SCSI_PutB2(scsiWr+sICR, 0);
   SCSI_PutB2(scsiRd+sMR, 0);
   SCSI_PutB2(scsiWr+sMR, 0);
   SCSI_PutB2(scsiRd+sTCR, 0);
   SCSI_PutB2(scsiWr+sTCR, 0);
   SCSI_PutB2(scsiRd+sCSR, 0);
   SCSI_PutB2(scsiWr+sSER, 0);
   SCSI_PutB2(scsiRd+sBSR, 0);
   SCSI_PutB2(scsiWr+sDMAtx, 0);
   SCSI_PutB2(scsiRd+sIDR, 0);
   SCSI_PutB2(scsiWr+sTDMArx, 0);
   SCSI_PutB2(scsiRd+sRESET, 0);
   SCSI_PutB2(scsiWr+sIDMArx, 0);
   SCSI_PutB2(scsiRd+sODR+dackWr, 0);
   SCSI_PutB2(scsiWr+sIDR+dackRd, 0);

   return;
}

void SCSI_BusReset (void)
{
   SCSI_PutB2(scsiRd+sCDR, 0);
   SCSI_PutB2(scsiWr+sODR, 0);
   SCSI_PutB2(scsiRd+sICR, SCSI_GetB2(scsiRd+sICR) & 0x80);
   SCSI_PutB2(scsiWr+sICR, SCSI_GetB2(scsiWr+sICR) & 0x80);
   SCSI_PutB2(scsiRd+sMR, SCSI_GetB2(scsiRd+sMR) & 0x40);
   SCSI_PutB2(scsiWr+sMR, SCSI_GetB2(scsiWr+sMR) & 0x40);
   SCSI_PutB2(scsiRd+sTCR, 0);
   SCSI_PutB2(scsiWr+sTCR, 0);
   SCSI_PutB2(scsiRd+sCSR, 0x80);
   SCSI_PutB2(scsiWr+sSER, 0);
   SCSI_PutB2(scsiRd+sBSR, 0x10);
   SCSI_PutB2(scsiWr+sDMAtx, 0);
   SCSI_PutB2(scsiRd+sIDR, 0);
   SCSI_PutB2(scsiWr+sTDMArx, 0);
   SCSI_PutB2(scsiRd+sRESET, 0);
   SCSI_PutB2(scsiWr+sIDMArx, 0);
   SCSI_PutB2(scsiRd+sODR+dackWr, 0);
   SCSI_PutB2(scsiWr+sIDR+dackRd, 0);

   SCSI_Chip.IRQ = 1;
   SCSI_Chip.DRQ = 0; 
   SCSI_Chip.EOP = 0;
   SCSI_Chip.DACK = 0;
   SCSI_Chip.READY = 0;
   SCSI_Chip.A0 = 0;
   SCSI_Chip.A1 = 0;
   SCSI_Chip.A2 = 0;
   SCSI_Chip.CS = 0;
   SCSI_Chip.I_OW = 0;
   SCSI_Chip.I_OR = 0;
   SCSI_Chip.D0_D7 = 0;
   SCSI_Chip.MSG = 0;
   SCSI_Chip.C_D = 0;
   SCSI_Chip.I_O = 0;
   SCSI_Chip.ACK = 0;
   SCSI_Chip.REQ = 0;
   SCSI_Chip.DBP = 0;
   SCSI_Chip.DB0_DB7 = 0;
   SCSI_Chip.BSY = 0;
   SCSI_Chip.SEL = 0;
   SCSI_Chip.ATN = 0;
   
   if (calls==0) {
      patch_globals();
      calls=1;
   }

   return;
}

void SCSI_CheckPins (void)
{
   if (SCSI_Chip.RESET==1) {
      SCSI_ChipReset();
   }

   if (SCSI_Chip.RST==1) {
      SCSI_PutB2(scsiRd+sICR, SCSI_GetB2(scsiRd+sICR) | 0x80);
      SCSI_PutB2(scsiRd+sCSR, 0x80);
      SCSI_PutB2(scsiRd+sBSR, 0x10);
      SCSI_BusReset();
   } else {
      SCSI_PutB2(scsiRd+sICR, SCSI_GetB2(scsiRd+sICR) & ~0x80);
      SCSI_PutB2(scsiRd+sCSR, SCSI_GetB2(scsiRd+sCSR) & ~0x80);
   }

   if (SCSI_Chip.SEL==1) {
      SCSI_PutB2(scsiRd+sCSR, SCSI_GetB2(scsiRd+sCSR) | 0x02);
      SCSI_PutB2(scsiRd+sBSR, 0x10);
   } else {
      SCSI_PutB2(scsiRd+sCSR, SCSI_GetB2(scsiRd+sCSR) & ~0x02);
   }   
}

void SCSI_Check (void)
{
   static UBYTE old1=1;    // just to check the previous state of the bit

      SCSI_CheckPins();

      // check the chip registers, AS SET BY THE CPU
      if ((SCSI_GetB2(scsiWr+sICR) >> 7) == 1) {  // Check Assert RST
         SCSI_Chip.RST=1;
      } else {
         SCSI_Chip.RST=0;
      }

      if ((SCSI_GetB2(scsiWr+sICR) >> 2) == 1) { // Check Assert SEL
         SCSI_Chip.SEL = 1;
      } else {
         SCSI_Chip.SEL = 0;
      }

      // The arbitration select/reselect scenario [stub.. doesn't really work...]
      if ((SCSI_GetB2(scsiWr+sODR) >> 7) == 1) {  // Check if the Mac tries to be an initiator
         if ((SCSI_GetB2(scsiWr+sMR) & 1) == 1) { // the Mac set arbitration in progress
            // stub! tell the mac that there is arbitration in progress...
            SCSI_PutB2(scsiRd+sICR, SCSI_GetB2(scsiRd+sICR) | 0x40);
            // ... that we didn't lose arbitration ...
            SCSI_PutB2(scsiRd+sICR, SCSI_GetB2(scsiRd+sICR) & ~0x20);
            // ... and that there isn't a higher priority ID present...
            SCSI_PutB2(scsiRd+sCDR, 0x00);

            // ... the arbitration and selection/reselection is complete.
            // the initiator tries to connect to the SCSI device, fails
            // and returns after timeout.
         }
      }

      SCSI_CheckPins();

   return;
}

FILE *fp;

UBYTE SCSI_GetB (CPTR addr)
{
//   fprintf(stderr, "%4lx: SCSI GB %4lx --> %2x\n", m68k_getpc(), addr, (SCSI[addr >> 1] >> (addr & 1 ? 0 : 8)));

   return SCSI [addr >> 1] >> (addr & 1 ? 0 : 8);
}

UWORD SCSI_GetW (CPTR addr)
{
   return SCSI [addr >> 1];
}

ULONG SCSI_GetL (CPTR addr)
{
   return ((ULONG) SCSI [addr >> 1] << 16) | SCSI [(addr >> 1)+1];
}

void SCSI_PutB (CPTR addr, UBYTE in)
{
//   fprintf(stderr, "%4lx: SCSI PB %4lx <-- %2x\n", m68k_getpc(), addr, in);
   CheckSCSI=1;

   if (!(addr & 1)) {
      SCSI [addr >> 1] = (SCSI [addr >> 1] & 0xff) | (((UWORD) in) << 8);
   } else {
      SCSI [addr >> 1] = (SCSI [addr >> 1] & 0xff00) | in;
   }

   return;
}

void SCSI_PutW (CPTR addr, UWORD in)
{
   SCSI [addr >> 1] = in;

   return;
}

void SCSI_PutL (CPTR addr, ULONG in)
{
   SCSI [addr >> 1] = in >> 16;
   SCSI [(addr >> 1) + 1] = (UWORD) in;

   return;
}

/* internal memory access routines */
UBYTE SCSI_GetB2 (CPTR addr)
{
   return SCSI [addr >> 1] >> (addr & 1 ? 0 : 8);
}

UWORD SCSI_GetW2 (CPTR addr)
{
   return SCSI [addr >> 1];
}

ULONG SCSI_GetL2 (CPTR addr)
{
   return ((ULONG) SCSI [addr >> 1] << 16) | SCSI [(addr >> 1)+1];
}

void SCSI_PutB2 (CPTR addr, UBYTE in)
{
   CheckSCSI=1;

   if (!(addr & 1)) {
      SCSI [addr >> 1] = (SCSI [addr >> 1] & 0xff) | (((UWORD) in) << 8);
   } else {
      SCSI [addr >> 1] = (SCSI [addr >> 1] & 0xff00) | in;
   }

   return;
}

void SCSI_PutW2 (CPTR addr, UWORD in)
{
   SCSI [addr >> 1] = in;

   return;
}

void SCSI_PutL2 (CPTR addr, ULONG in)
{
   SCSI [addr >> 1] = in >> 16;
   SCSI [(addr >> 1) + 1] = (UWORD) in;

   return;
}
