/*
   vMac 0.1 Source
   By Philip Cummins (PC)
   Copyright PC 1997
*/

#include "sysconfi.h"                   // includes/sysconfig.h
#include "sysdeps.h"
#include "config.h"
#include "options.h"

#include "prefs.h"
#include "memory2.h"                    // includes/memory.h

#include "custom2.h"                    // includes/vMac_custom.h
#include "readcpu.h"
#include "newcpu.h"
#include "debug.h"

#include "Memory.h"
#include "Sony.h"
#include "Screen.h"
#include "Trap.h"
#include "Driver.h"

#include "IBM-SND.h"
#include "IBM-VGA.h"

#include <dpmi.h>
#include <go32.h>
#include <stdlib.h>

int Debugger=0;
int Install=1;
int use_debugger = 0;
char ConfigFile[] = "vMac.cfg";
extern void debug(void);

void Quit_vMac(void);
void Help(void);
void Parse(int argc, char **argv);

void Quit_vMac(void)
{
   extern void allegro_exit (void);

   PrefsSave(ConfigFile);
   fprintf(stderr, "vMac: quit ok.\nvMac: %s saved.\n", ConfigFile);

   if (thePrefs.Sound) Sound_Uninstall();
   UninstallExceptionHandler();
   allegro_exit();
   exit(0);
}  

void Help (void)
{
   printf("vMac: The Virtual Macintosh Emulator.    \n");
   printf("      version %s for MS-DOS systems.\n\n", VMACVERSION);
   printf("Summary of command-line options:         \n");
   printf("  -h, --help              : Print help   \n");
   printf("  -d, --debugger          : Start vMac debugger\n");
   printf("  -c, --config [filename] : Use a different file for config instead of vMac.cfg\n");
}

void Parse (int argc, char **argv)
{
   int i;
   
   for (i=1; i<argc; i++) {
      if ((!strcmp(argv[i], "-h")) || (!strcmp(argv[i], "--help"))) {
         Help();
         exit(0);
      }
      else if ((!strcmp(argv[i], "-c")) || (!strcmp(argv[i], "--config"))) {
         i++;
         strcpy(ConfigFile, argv[i]);
      }
      else if ((!strcmp(argv[i], "-d")) || (!strcmp(argv[i], "--debugger"))) {
         i++;
         Debugger=1;
      }
      else {
         printf("vMac: unrecognized: %s\n", argv[i]);
         exit(0);
      }
   }
}

// vMac/DOS 0.1.9.5 has severe stability problems when run under Windows.
// This detects if windows is running in the background, and kills it
// gracefully.
void Kill_Windows (void)
{
   __dpmi_regs Regs;
   int detected=0;

   // Detect Win 3.x
   Regs.x.ax=0x160a;
   __dpmi_int(0x2f, &Regs);
   if (Regs.x.ax==0)
      detected=1;

   // Normally, the previous algorithm should've worked for any version
   // of Windows, including Windows95.  However, Windows95 can make itself
   // "invisible"... That's why we get the Virtual Machine Manager API
   // entry-point, for a fail-safe detection of Windows95/98/NT
   Regs.x.ax = 0x1684;                                  
   Regs.x.bx = 0x0001;
   Regs.x.di = 0x0000;
   Regs.x.es = 0x0000;
   __dpmi_int(0x2f, &Regs);
   if (Regs.x.es!=0 || Regs.x.di!=0)
      detected=1;

   if (detected) {
      printf("vMac/DOS must be run in MS-DOS mode.  Sorry. \n");
      exit(0);
   }

   return;
}

int main(int argc, char **argv)
{
   extern int allegro_init(void);

   printf("vMac: The Virtual Macintosh Emulator.  \n");
   printf("      version %s for MS-DOS systems. \n\n", VMACVERSION);

   Kill_Windows();
   Parse(argc, argv);

   PrefsLoad(ConfigFile);

   allegro_init();
   InstallExceptionHandler();

   if (custominit() == 0) {
      m68k_reset();

      if (thePrefs.Sound) {
         if (Sound_Install())
            Quit_vMac();
      }

      init_m68k();
      Driver_LockDrivers();
      Screen_Init(argc, argv);

      if (Debugger)
         use_debugger=1;
      else
         use_debugger=0;

      debug();
   } else {
      fprintf(stderr, "vMac: Sorry, errors encountered - cannot continue.\n");
   }

   return 0;
}
